@@ -10,8 +10,9 @@ import pickle
 import io
 from abc import ABC, abstractmethod
 import pandas
+import json
 
-from vantage6.tools.dispatch_rpc import dispact_rpc
+from vantage6.tools.dispatch_rpc import dispatch_rpc
 from vantage6.tools.util import info
 from vantage6.tools import deserialization, serialization
 from vantage6.tools.data_format import DataFormat
@@ -25,9 +26,9 @@ _MAX_FORMAT_STRING_LENGTH = 10
 _SPARQL_RETURN_FORMAT = CSV
 
 
-def docker_wrapper(module: str):
+def docker_wrapper(module: str, load_data=True):
     wrapper = DockerWrapper()
-    wrapper.wrap_algorithm(module)
+    wrapper.wrap_algorithm(module, load_data)
 
 
 def sparql_wrapper(module: str):
@@ -35,46 +36,58 @@ def sparql_wrapper(module: str):
     wrapper.wrap_algorithm(module)
 
 
+def parquet_wrapper(module: str):
+    wrapper = ParquetWrapper()
+    wrapper.wrap_algorithm(module)
+
+
+def multidb_wrapper(module: str):
+    wrapper = MultiDBWrapper()
+    wrapper.wrap_algorithm(module)
+
+
 class WrapperBase(ABC):
 
-    def wrap_algorithm(self, module):
+    def wrap_algorithm(self, module, load_data=True):
         """
-            Wrap an algorithm module to provide input and output handling for the
-            vantage6 infrastructure.
+        Wrap an algorithm module to provide input and output handling for the
+        vantage6 infrastructure.
 
-            Data is received in the form of files, whose location should be specified
-            in the following environment variables:
-            - `INPUT_FILE`: input arguments for the algorithm
-            - `OUTPUT_FILE`: location where the results of the algorithm should be
-              stored
-            - `TOKEN_FILE`: access token for the vantage6 server REST api
-            - `DATABASE_URI`: either a database endpoint or path to a csv file.
+        Data is received in the form of files, whose location should be
+        specified in the following environment variables:
+        - `INPUT_FILE`: input arguments for the algorithm
+        - `OUTPUT_FILE`: location where the results of the algorithm should be
+            stored
+        - `TOKEN_FILE`: access token for the vantage6 server REST api
+        - `DATABASE_URI`: either a database endpoint or path to a csv file.
 
-            The wrapper is able to parse a number of input file formats. The available
-            formats can be found in `vantage6.tools.data_format.DataFormat`. When the
-            input is not pickle (legacy), the format should be specified in the first
-            bytes of the input file, followed by a '.'.
+        The wrapper is able to parse a number of input file formats. The
+        available formats can be found in
+        `vantage6.tools.data_format.DataFormat`. When the input is not pickle
+        (legacy), the format should be specified in the first bytes of the
+        input file, followed by a '.'.
 
-            It is also possible to specify the desired output format. This is done by
-            including the parameter 'output_format' in the input parameters. Again, the
-            list of possible output formats can be found in
-            `vantage6.tools.data_format.DataFormat`.
+        It is also possible to specify the desired output format. This is done
+        by including the parameter 'output_format' in the input parameters.
+        Again, the list of possible output formats can be found in
+        `vantage6.tools.data_format.DataFormat`.
 
-            It is still possible that output serialization will fail even if the
-            specified format is listed in the DataFormat enum. Algorithms can in
-            principle return any python object, but not every serialization format will
-            support arbitrary python objects. When dealing with unsupported algorithm
-            output, the user should use 'pickle' as output format, which is the
-            default.
+        It is still possible that output serialization will fail even if the
+        specified format is listed in the DataFormat enum. Algorithms can in
+        principle return any python object, but not every serialization format
+        will support arbitrary python objects. When dealing with unsupported
+        algorithm output, the user should use 'pickle' as output format, which
+        is the default.
 
-            The other serialization formats support the following algorithm output:
-            - built-in primitives (int, float, str, etc.)
-            - built-in collections (list, dict, tuple, etc.)
-            - pandas DataFrames
+        The other serialization formats support the following algorithm output:
+        - built-in primitives (int, float, str, etc.)
+        - built-in collections (list, dict, tuple, etc.)
+        - pandas DataFrames
 
-            :param module: module that contains the vantage6 algorithms
-            :return:
-            """
+        :param module: module that contains the vantage6 algorithms
+        :param load_data: attempt to load the data or execute the query
+        :return:
+        """
         info(f"wrapper for {module}")
 
         # read input from the mounted inputfile.
@@ -91,17 +104,22 @@ class WrapperBase(ABC):
         with open(token_file) as fp:
             token = fp.read().strip()
 
+        # TODO in v4+, we should work with multiple databases instead of this
+        # default one
         database_uri = os.environ["DATABASE_URI"]
         info(f"Using '{database_uri}' as database")
-        # with open(data_file, "r") as fp:
-        data = self.load_data(database_uri, input_data)
+
+        if load_data:
+            data = self.load_data(database_uri, input_data)
+        else:
+            data = None
 
         # make the actual call to the method/function
         info("Dispatching ...")
-        output = dispact_rpc(data, input_data, module, token)
+        output = dispatch_rpc(data, input_data, module, token)
 
         # write output from the method to mounted output file. Which will be
-        # transfered back to the server by the node-instance.
+        # transferred back to the server by the node-instance.
         output_file = os.environ["OUTPUT_FILE"]
         info(f"Writing output to {output_file}")
 
@@ -120,6 +138,9 @@ class DockerWrapper(WrapperBase):
         return pandas.read_csv(database_uri)
 
 
+CsvWrapper = DockerWrapper
+
+
 class SparqlDockerWrapper(WrapperBase):
     @staticmethod
     def load_data(database_uri, input_data):
@@ -136,11 +157,29 @@ class SparqlDockerWrapper(WrapperBase):
         return pandas.read_csv(io.StringIO(result))
 
 
+class ParquetWrapper(WrapperBase):
+    @staticmethod
+    def load_data(database_uri, input_data):
+        return pandas.read_parquet(database_uri)
+
+
+class MultiDBWrapper(WrapperBase):
+    @staticmethod
+    def load_data(database_uri, input_data):
+        db_labels = json.loads(os.environ.get("DB_LABELS"))
+        databases = {}
+        for db_label in db_labels:
+            db_env_var = f'{db_label.upper()}_DATABASE_URI'
+            databases[db_label] = os.environ.get(db_env_var)
+        return databases
+
+
 def write_output(output_format, output, output_file):
     """
     Write output to output_file using the format from output_format.
 
-    If output_format == None, write output as pickle without indicating format (legacy method)
+    If output_format == None, write output as pickle without indicating format
+    (legacy method)
 
     :param output_format:
     :param output:
